/* (C) 1999-2001 Paul `Rusty' Russell
 * (C) 2002-2006 Netfilter Core Team <coreteam@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Everything about the rules for NAT. */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <net/route.h>
#include <linux/bitops.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <net/netfilter/nf_nat.h>
#include <net/netfilter/nf_nat_core.h>
#include <net/netfilter/nf_nat_rule.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define NAT_VALID_HOOKS ((1<<NF_IP_PRE_ROUTING) | (1<<NF_IP_POST_ROUTING) | (1<<NF_IP_LOCAL_OUT))
#ifdef TEAMF1_CHANGES
#include<linux/netfilter_ipv4/lockhelp.h>
#include<linux/netdevice.h>
#include<linux/inetdevice.h> 
#define TF1_DBG 1000
/* Lock protects snat region inside conntrack */
static DECLARE_RWLOCK(snat_lock);
#define SIP_PORT 5060
#define TEAMF1_OTO_IB_NAT_CHANGES
#endif /* TEAMF1_CHANGES */

/* Enable Strict One to One Inbound Nat */
int  strictDnat = 1;

static struct
{
	struct ipt_replace repl;
	struct ipt_standard entries[3];
	struct ipt_error term;
} nat_initial_table __initdata = {
	.repl = {
		.name = "nat",
		.valid_hooks = NAT_VALID_HOOKS,
		.num_entries = 4,
		.size = sizeof(struct ipt_standard) * 3 + sizeof(struct ipt_error),
		.hook_entry = {
			[NF_IP_PRE_ROUTING] = 0,
			[NF_IP_POST_ROUTING] = sizeof(struct ipt_standard),
			[NF_IP_LOCAL_OUT] = sizeof(struct ipt_standard) * 2 },
		.underflow = {
			[NF_IP_PRE_ROUTING] = 0,
			[NF_IP_POST_ROUTING] = sizeof(struct ipt_standard),
			[NF_IP_LOCAL_OUT] = sizeof(struct ipt_standard) * 2 },
	},
	.entries = {
		/* PRE_ROUTING */
		{
			.entry = {
				.target_offset = sizeof(struct ipt_entry),
				.next_offset = sizeof(struct ipt_standard),
			},
			.target = {
				.target = {
					.u = {
						.target_size = IPT_ALIGN(sizeof(struct ipt_standard_target)),
					},
				},
				.verdict = -NF_ACCEPT - 1,
			},
		},
		/* POST_ROUTING */
		{
			.entry = {
				.target_offset = sizeof(struct ipt_entry),
				.next_offset = sizeof(struct ipt_standard),
			},
			.target = {
				.target = {
					.u = {
						.target_size = IPT_ALIGN(sizeof(struct ipt_standard_target)),
					},
				},
				.verdict = -NF_ACCEPT - 1,
			},
		},
		/* LOCAL_OUT */
		{
			.entry = {
				.target_offset = sizeof(struct ipt_entry),
				.next_offset = sizeof(struct ipt_standard),
			},
			.target = {
				.target = {
					.u = {
						.target_size = IPT_ALIGN(sizeof(struct ipt_standard_target)),
					},
				},
				.verdict = -NF_ACCEPT - 1,
			},
		},
	},
	/* ERROR */
	.term = {
		.entry = {
			.target_offset = sizeof(struct ipt_entry),
			.next_offset = sizeof(struct ipt_error),
		},
		.target = {
			.target = {
				.u = {
					.user = {
						.target_size = IPT_ALIGN(sizeof(struct ipt_error_target)),
						.name = IPT_ERROR_TARGET,
					},
				},
			},
			.errorname = "ERROR",
		},
	}
};

static struct xt_table nat_table = {
	.name		= "nat",
#ifdef TEAMF1_CHANGES
	.table		= &nat_initial_table.repl,
#endif /* TEAMF1_CHANGES */
	.valid_hooks	= NAT_VALID_HOOKS,
	.lock		= RW_LOCK_UNLOCKED,
	.me		= THIS_MODULE,
	.af		= AF_INET,
};

/* Source NAT */
static unsigned int ipt_snat_target(struct sk_buff **pskb,
				    const struct net_device *in,
				    const struct net_device *out,
				    unsigned int hooknum,
				    const struct xt_target *target,
				    const void *targinfo)
{
	struct nf_conn *ct;
	enum ip_conntrack_info ctinfo;
	const struct nf_nat_multi_range_compat *mr = targinfo;

	NF_CT_ASSERT(hooknum == NF_IP_POST_ROUTING);

	ct = nf_ct_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	NF_CT_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED ||
			    ctinfo == IP_CT_RELATED + IP_CT_IS_REPLY));
	NF_CT_ASSERT(out);
#ifdef TEAMF1_CHANGES
	struct nf_conn_nat *nat = nfct_nat(ct);
	//WRITE_LOCK(&snat_lock);
	nat->snat_index = out->ifindex;
	//WRITE_UNLOCK(&snat_lock);
#endif /* TEAMF1_CHANGES */
	return nf_nat_setup_info(ct, &mr->range[0], hooknum);
}

/* Before 2.6.11 we did implicit source NAT if required. Warn about change. */
static void warn_if_extra_mangle(__be32 dstip, __be32 srcip)
{
	static int warned = 0;
	struct flowi fl = { .nl_u = { .ip4_u = { .daddr = dstip } } };
	struct rtable *rt;

	if (ip_route_output_key(&rt, &fl) != 0)
		return;

	if (rt->rt_src != srcip && !warned) {
		DEBUGP("NAT: no longer support implicit source local NAT\n");
		DEBUGP("NAT: packet src %u.%u.%u.%u -> dst %u.%u.%u.%u\n",
		       NIPQUAD(srcip), NIPQUAD(dstip));
		warned = 1;
	}
	ip_rt_put(rt);
}

static unsigned int ipt_dnat_target(struct sk_buff **pskb,
				    const struct net_device *in,
				    const struct net_device *out,
				    unsigned int hooknum,
				    const struct xt_target *target,
				    const void *targinfo)
{
	struct nf_conn *ct;
	enum ip_conntrack_info ctinfo;
	const struct nf_nat_multi_range_compat *mr = targinfo;
    unsigned int ret = 0;

	NF_CT_ASSERT(hooknum == NF_IP_PRE_ROUTING ||
		     hooknum == NF_IP_LOCAL_OUT);

	ct = nf_ct_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	NF_CT_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));
#ifdef TEAMF1_CHANGES
#define FW_DNAT_CONNMARK            0x2222   /* DNAT'd connections */ 
#define FW_DNAT_CONNMARK_MASK       0x300   /* 9,10 bit represent DNAT */    
        /* This mark is set to identify all the dnat'd packets, so that
         * the dnat'ed packets can be bypassed in the filter table with
         * out much examination. 
         */
        ct->mark = FW_DNAT_CONNMARK;
#endif /* TEAMF1_CHANGES */

	if (hooknum == NF_IP_LOCAL_OUT &&
	    mr->range[0].flags & IP_NAT_RANGE_MAP_IPS)
		warn_if_extra_mangle((*pskb)->nh.iph->daddr,
				     mr->range[0].min_ip);
#ifdef TEAMF1_OTO_IB_NAT_CHANGES
        if (strictDnat)
        if(mr->otoNatOrgDstMinIp && mr->otoNatOrgDstMaxIp) {
            /* Get the source address and calculate the IP to DNAT to */
            u_int32_t dstIp;
            u_int32_t offset;
            dstIp = ct->tuplehash[0].tuple.dst.u3.ip;

            /* We do not know how we reached here, its fate be decided
             * un-modified.
             */
            if (dstIp > mr->otoNatOrgDstMaxIp ||
                dstIp < mr->otoNatOrgDstMinIp)
                return IPT_CONTINUE;  

            offset = dstIp - mr->otoNatOrgDstMinIp;

            struct nf_nat_multi_range_compat *strictRangeCompat = NULL;

            strictRangeCompat = kmalloc(sizeof
                                       (struct nf_nat_multi_range_compat),
                                        GFP_KERNEL);
            if (!strictRangeCompat) {
                DEBUGP(KERN_WARNING "%s:No memory !\n", __FUNCTION__);
                return 0;
            }
            memcpy(strictRangeCompat, mr, sizeof(struct
                                          nf_nat_multi_range_compat));
            /* We need to reset max_ip here, otherwise ip_nat_setup_info
             * randomly choose an IP between min_ip and max_ip
             */

            strictRangeCompat->range[0].min_ip += offset;
            strictRangeCompat->range[0].max_ip = 
                strictRangeCompat->range[0].min_ip;
            

            DEBUGP("%u.%u.%u.%u OTO IB NAT to %u.%u.%u.%u..",
                   NIPQUAD(dstIp),
                   NIPQUAD(strictRangeCompat->range[0].min_ip));
            /* TODO: OTO IB NAT on range ports */
	        ret = nf_nat_setup_info(ct, &strictRangeCompat->range[0], hooknum);
            kfree(strictRangeCompat);
            return ret;
        }
#endif
defaultNat:
	return nf_nat_setup_info(ct, &mr->range[0], hooknum);
}

static int ipt_snat_checkentry(const char *tablename,
			       const void *entry,
			       const struct xt_target *target,
			       void *targinfo,
			       unsigned int hook_mask)
{
	struct nf_nat_multi_range_compat *mr = targinfo;

	/* Must be a valid range */
	if (mr->rangesize != 1) {
		DEBUGP("SNAT: multiple ranges no longer supported\n");
		return 0;
	}
	return 1;
}

static int ipt_dnat_checkentry(const char *tablename,
			       const void *entry,
			       const struct xt_target *target,
			       void *targinfo,
			       unsigned int hook_mask)
{
	struct nf_nat_multi_range_compat *mr = targinfo;

	/* Must be a valid range */
	if (mr->rangesize != 1) {
		DEBUGP("DNAT: multiple ranges no longer supported\n");
		return 0;
	}
	if (mr->range[0].flags & IP_NAT_RANGE_PROTO_RANDOM) {
		DEBUGP("DNAT: port randomization not supported\n");
		return 0;
	}
	return 1;
}

inline unsigned int
alloc_null_binding(struct nf_conn *ct,
		   struct nf_nat_info *info,
		   unsigned int hooknum)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED).
	   Use reply in case it's already been mangled (eg local packet).
	*/
	__be32 ip
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u3.ip
		   : ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u3.ip);
	struct nf_nat_range range
		= { IP_NAT_RANGE_MAP_IPS, ip, ip, { 0 }, { 0 } };

	DEBUGP("Allocating NULL binding for %p (%u.%u.%u.%u)\n",
	       ct, NIPQUAD(ip));
	return nf_nat_setup_info(ct, &range, hooknum);
}

unsigned int
alloc_null_binding_confirmed(struct nf_conn *ct,
			     struct nf_nat_info *info,
			     unsigned int hooknum)
{
	__be32 ip
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u3.ip
		   : ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u3.ip);
	u_int16_t all
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.all
		   : ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.all);
	struct nf_nat_range range
		= { IP_NAT_RANGE_MAP_IPS, ip, ip, { all }, { all } };

	DEBUGP("Allocating NULL binding for confirmed %p (%u.%u.%u.%u)\n",
	       ct, NIPQUAD(ip));
	return nf_nat_setup_info(ct, &range, hooknum);
}

int nf_nat_rule_find(struct sk_buff **pskb,
		     unsigned int hooknum,
		     const struct net_device *in,
		     const struct net_device *out,
		     struct nf_conn *ct,
		     struct nf_nat_info *info)
{
	int ret;

	ret = ipt_do_table(pskb, hooknum, in, out, &nat_table);

	if (ret == NF_ACCEPT) {
		if (!nf_nat_initialized(ct, HOOK2MANIP(hooknum)))
			/* NUL mapping */
			ret = alloc_null_binding(ct, info, hooknum);
	}
	return ret;
}

#ifdef TEAMF1_CHANGES
static inline int
device_cmp(struct nf_conn *i, void *ifindex)
{
	int ret;

	struct nf_conn_nat *nat = nfct_nat(i);
//	READ_LOCK(&snat_lock);
    if (!nat)
        return 0;

	ret = (nat->snat_index == (int)(long)ifindex);
//	READ_UNLOCK(&snat_lock);

	return ret;
}

static inline int
sip_port_cleanup(struct nf_conn *i, void *port)
{
	int ret;
	read_lock_bh(&snat_lock);
	ret = (i->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port == __constant_htons(SIP_PORT));
	read_unlock_bh(&snat_lock);
	return ret;

}
static int snat_device_event(struct notifier_block *this,
			     unsigned long event,
			     void *ptr)
{
	struct net_device *dev = ptr;

	if (event == NETDEV_DOWN) {
		/* Device was downed.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		IP_NF_ASSERT(dev->ifindex != 0);

		nf_ct_iterate_cleanup(device_cmp, (void *)(long)dev->ifindex);
	}
	if (event == NETDEV_GENERIC) {
		DEBUGP(KERN_DEBUG "%s: Cleaning sip entries\n",__func__);
		/* Clean sip related conntracks */
		nf_ct_iterate_cleanup(sip_port_cleanup, NULL);
	}

	return NOTIFY_DONE;
}

static int snat_inet_event(struct notifier_block *this,
			   unsigned long event,
			   void *ptr)
{
	struct net_device *dev = ((struct in_ifaddr *)ptr)->ifa_dev->dev;

	if (event == NETDEV_DOWN) {
		/* IP address was deleted.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		IP_NF_ASSERT(dev->ifindex != 0);

		nf_ct_iterate_cleanup(device_cmp, (void *)(long)dev->ifindex);
	}

	return NOTIFY_DONE;
}

static struct notifier_block snat_dev_notifier = {
	.notifier_call	= snat_device_event,
};

static struct notifier_block snat_inet_notifier = {
	.notifier_call	= snat_inet_event,
};
#endif /* TEAMF1_CHANGES */

static struct xt_target ipt_snat_reg = {
	.name		= "SNAT",
	.target		= ipt_snat_target,
	.targetsize	= sizeof(struct nf_nat_multi_range_compat),
	.table		= "nat",
	.hooks		= 1 << NF_IP_POST_ROUTING,
	.checkentry	= ipt_snat_checkentry,
	.family		= AF_INET,
};

static struct xt_target ipt_dnat_reg = {
	.name		= "DNAT",
	.target		= ipt_dnat_target,
	.targetsize	= sizeof(struct nf_nat_multi_range_compat),
	.table		= "nat",
	.hooks		= (1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_LOCAL_OUT),
	.checkentry	= ipt_dnat_checkentry,
	.family		= AF_INET,
};

#ifdef TEAMF1_OTO_IB_NAT_CHANGES
/*  creates /proc/sys/tf1/strictDnat, to enable/disable OTO IB NAT. */
static ctl_table strict_dnat_table[] = {
    { .ctl_name     = TF1_DBG + 2,  
      .procname     = "strictDnat",
      .data         = &strictDnat,
      .maxlen       = sizeof(strictDnat),
      .mode         = 0644,
      .proc_handler = &proc_dointvec 
    },
    {}
};

/*  This adds a child to /proc/sys/tf1 */
static ctl_table strict_dnat_root_table[] = {
    { .ctl_name    = TF1_DBG,
      .procname    = "tf1",
      .mode        = 0555,
      .child       = strict_dnat_table },
    {}
};
#endif
int __init nf_nat_rule_init(void)
{
	int ret;

	ret = ipt_register_table(&nat_table, &nat_initial_table.repl);
	if (ret != 0)
		return ret;
	ret = xt_register_target(&ipt_snat_reg);
	if (ret != 0)
		goto unregister_table;

	ret = xt_register_target(&ipt_dnat_reg);
	if (ret != 0)
		goto unregister_snat;

#ifdef TEAMF1_CHANGES
	/* Register for device down reports */
    	register_netdevice_notifier(&snat_dev_notifier);

	/* Register IP address change reports */
	register_inetaddr_notifier(&snat_inet_notifier);
#endif /* TEAMF1_CHANGES */
#ifdef TEAMF1_OTO_IB_NAT_CHANGES
    register_sysctl_table(strict_dnat_root_table); 
#endif /* TEAMF1_OTO_IB_NAT_CHANGES */
	return ret;

 unregister_snat:
	xt_unregister_target(&ipt_snat_reg);
 unregister_table:
	ipt_unregister_table(&nat_table);

	return ret;
}

void nf_nat_rule_cleanup(void)
{
	xt_unregister_target(&ipt_dnat_reg);
	xt_unregister_target(&ipt_snat_reg);
	ipt_unregister_table(&nat_table);
#ifdef TEAMF1_CHANGES
	unregister_netdevice_notifier(&snat_dev_notifier);
	unregister_inetaddr_notifier(&snat_inet_notifier);    
#endif /* TEAMF1_CHANGES */    
#ifdef TEAMF1_OTO_IB_NAT_CHANGES
    unregister_sysctl_table((struct ctl_table_header *)&strict_dnat_root_table);
#endif /* TEAMF1_OTO_IB_NAT_CHANGES */
}
